#
# common BIND definitions
#
# Jason Vas Dias <jvdias@redhat.com>, Red Hat Inc. Sep 2004

FORWARD=1

class Data: # "Abstract" base class for DNS Data Types
    def check(this, str):
        return False
    
    def __str__(this): # should return fully qualified name or address 
        return this.value()

    def value(this):   # should return input used to construct 
        return this.__str__()

    def label(this):   # should return GUI label / human readable representation
        return this.__str__()

    def relative_label(this): # should return GUI label / human readable representation
                              # relative to "$ORIGIN"
        return this.label()

    def relative_value(this,origin=None): # should return input format value relative to origin
        return this.value()

    def out(this): # should return output string to be stored in configuration file
        return this.value()

        
import IPV4
import IPV6
import NSAP

import string

def is_absolute(s):
    return ( (len(s) > 0) and (s[-1] == '.') and ((len(s) < 2) or (s[-2] != '\\')) )

def make_absolute(s):
    if (len(s)<=0) or (s[-1] != '.'):
        s += '.'
    return s

def split_dots(s):
    a = []
    p = None
    t = ''
    for c in s:
        if ( p != '\\' ) and (c == '.'):
            a.append(t)
            t=''
        else:
            t += c
        p = c
    if len(t):
        a.append(t)
    return a


class Name(Data):
    data=''
    origin=None
    SRV=False
    
    def __init__(self, name, zone=None, SRV=False):
        self.data = name
        if( zone != None ):
            self.origin = zone.origin           # origin at time of construction ($ORIGIN!)
            self.zone_origin = zone.zone_origin # zone's named.conf origin
            self.SRV = SRV
    
    def name(self):
        if (self.data==None) or (len(self.data) == 0) or (self.data == '@') :
            return self.origin
        return self.data
    
    def value(self):
        return self.data

    def relative_value(self, origin=None):
        if origin == None:
            if has_attr(self,'zone_origin'):
                origin = self.zone_origin
            else:
                origin = ''
        s = self.__str__()
        if (len(origin)>0) and (len(s) >= len(origin)) and (s[-len(origin):]==origin):
            s=s[:-len(origin)]
            if( len(s) == 0):
                s = '@'
            if(len(s) > 1) and (s[-1] == '.'):
                s=s[:-1]
        return s
    
    def __str__(self):
        if self.data == None:
            return ''
        if (len(self.data) and is_absolute(self.data)):
            return self.data
        else:
            if (len(self.data) == 0) or (self.data == '@') :
                return self.origin
            elif len(self.data) and (self.origin != None) and len(self.origin) :
                if self.origin[0] == '.':
                    return self.data + self.origin
                else:
                    return self.data + '.' + self.origin
            else:
                return self.data
    
    def label(self):
        s = self.__str__()
        if (len(s)>1) and (s[-1] == '.' ):
            return s[:-1]
        return s
    
    def relative_label(self):
        s = self.__str__()
        if ( hasattr(self,'zone_origin') and len(s) >= len(self.zone_origin)):
            s=s[:-len(self.zone_origin)]
            if is_absolute(s):
               return s[:-1]
        return s    
    
    def check(self, n=None):
        if (n == None):
            n = self.__str__()
        if(len(n)==0):
            return [(-1,_("domain name is of zero length"))]
        if (len(n) > 254):
            return [(-1,_("domain name too long - it is over the 254 character limit"))]
        if (n[-1] != '.') or ((len(n)>=2) and (n[-2] == '\\')):                
            if ( self.origin != '.' ): 
                n += '.' + self.origin
            else:
                n += self.origin
        if n == self.origin:
            return None
        a=self.split_dots(n)
        i=0
        allowed_chars = string.digits + string.ascii_lowercase + string.ascii_uppercase + '-*' 
        allowed_first_chars =  string.ascii_lowercase + string.ascii_uppercase + string.digits + '*' 
        warnings=[]
        for l in a:
            i+=1
            if (len(l) == 0) and (i != len(a)):
                return [(-1,_("null label") + ' '+ str(i) + ' '+ _("not allowed"))]
            if (len(l) == 0):
                continue
            if (len(l) > 63):
                return [(-1,_("label") + ' '  + str(i) + ' ' + _("too long - it is over the 63 character limit"))]
            if self.SRV and i <= 2:
                if l[0] not in '_':
                    warnings.append([0,_("label") + ' ' + str(i) + ' : ' + l + "\n" + _("missing underscore - SRV RR should be in format _Service._Proto.Name (RFC 2782)")])
            else:
                if l[0] not in allowed_first_chars:
                    warnings.append([0,_("label") + ' ' + str(i) + ' : ' + l + "\n" + _("domain name labels should begin with a letter or number (RFC 1123)")])
            for c in l[1:]:
                if c not in allowed_chars:
                    warnings.append([0,_("label") + ' ' + str(i) + ' : ' + l + "\n" + _("domain name labels should contain only characters in the ranges A-Z,1-9 or '-' (RFC 1123)")])
        if len(warnings):
            return warnings
        return None
    
    def split_dots(self,s=None):
        if( s == None):
            s = self.__str__()
        return split_dots(s)
    
    def split_domain(s, n=None):
        if(n == None):
            n = str(s)
        a=split_dots(n)
        return (a[0],make_absolute(string.join(a[1:],'.')))

class TTL(Data):
    seconds=None
    not_seconds=False
    
    def __init__( self, token ):
        if (token==None) or (len(token) == 0):
            return
        self.input(token)

    def input(self,token):
        days=0
        weeks=0
        hours=0
        minutes=0
        sec=0
        self.seconds=None
        self.not_seconds=False
        n=None
        for c in token[:]:
            C=c.upper()
            if C == 'W':
                if (n==None) or (len(n) == 0):
                    n=None
                    break
                weeks=int(n)
                n=''
                self.not_seconds=True
            elif C == 'D':
                if (n==None) or (len(n) == 0):
                    n=None
                    break
                days=int(n)
                n=''
                self.not_seconds=True
            elif C == 'H':
                if (n==None) or (len(n) == 0):
                    n=None
                    break
                hours=int(n)
                n=''
                self.not_seconds=True
            elif C == 'M':
                if (n==None) or (len(n) == 0):
                    n=None
                    break
                minutes=int(n)
                n=''
                self.not_seconds=True
            elif C == 'S':
                if (n==None) or (len(n) == 0):
                    n=None
                    break
                sec=int(n)
                n=''
                self.not_seconds=True
            elif c in string.digits:
                if n == None:
                    n=c
                else:
                    n+=c
            else:
                n = None
                break
        if(n != None) and ((self.not_seconds==False) or (len(n)==0)):
            if self.not_seconds or (len(n)==0):
                n='0'
            self.seconds = (weeks * 604800) + (days * 86400) + (hours * 3600) + (minutes * 60) + sec + int(n);
        
    def __str__(self):
        if self.seconds == None:
            return ''
        if not self.not_seconds:
            return str(self.seconds)
        else:
            n=self.seconds
            s=''
            P=[['W', 604800],['D',86400],['H',3600],['M',60], ['S',1]]
            while (n > 0) and len(P) :
                p = P.pop(0)
                q = n / p[1]
                if q > 0:
                   s += str(q)+p[0]
                   n = n % p[1]
            if s == '':
                s='0'
            return s

    def wdhms(self):
        t = {'W':0,'D':0,'H':0,'M':0,'S':0}
        if self.seconds == None:
            return t
        else:
            n=self.seconds
            s=''
            P=[['W', 604800],['D',86400],['H',3600],['M',60], ['S',1]]
            while (n > 0) and len(P) :
                p = P.pop(0)
                q = n / p[1]
                if q > 0:
                   t[p[0]] = q
                   n = n % p[1]
            return t
    
    def label(self):
        fmt = self.not_seconds
        self.not_seconds = True
        s = self.__str__()
        self.not_seconds=fmt
        h=''
        n=''
        for c in s:
            C = c.upper()
            if C=='W':
                if int(n)>1:
                    h+=n+_(' weeks ')
                else:
                    h+=n+_(' week ')
                n=''
            elif C=='D':
                if int(n)>1:
                    h+=n+_(' days ')
                else:
                    h+=n+_(' day ')
                n=''
            elif C=='H':
                if int(n)>1:
                    h+=n+_(' hours ')
                else:
                    h+=n+_(' hour ')
                n=''
            elif C=='M':
                if int(n)>1:
                    h+=n+_(' minutes ')
                else:
                    h+=n+_(' minute ')
                n=''
            elif C=='S':
                if int(n)>1:
                    h+=n+_(' seconds ')
                else:
                    h+=n+_(' second ')
                n=''
            elif c in string.digits:
                n+=c
        return h

    def check(self, str):
        ttl = TTL(str)
        return (ttl.seconds != None)


class Email(Name):
    
    def emailLabel(s,new=None):
        if new == None:
            a=Name.split_dots(s)
        else:
            a=Name.split_dots(s,new)
        return a[0] + '@' + string.join(a[1:],'.') +'.'

    def dnsLabel(s,new=None):
        n = new
        if( new == None):
            n = Name.__str__(s)
        if( n.find('@') > -1 ):
            p=None
            done=False
            nn=''
            for c in n:
                if not(done) and (c == '@') and (p != '\\'):
                    c='.'
                    done=True
                nn+=c
                p=c
            n=nn                
        return n        
        
    def label(s):        
        return s.emailLabel()

    def relative_label(s):
        return s.label()
    
class Location(Data):
    #   The LOC record is expressed in a master file in the following format:
    #   <owner> <TTL> <class> LOC ( d1 [m1 [s1]] {"N"|"S"} d2 [m2 [s2]]
    #                               {"E"|"W"} alt["m"] [siz["m"] [hp["m"]
    #                               [vp["m"]]]] )
    #     where:
    #        d1:     [0 .. 90]            (degrees latitude)
    #        d2:     [0 .. 180]           (degrees longitude)
    #        m1, m2: [0 .. 59]            (minutes latitude/longitude)
    #        s1, s2: [0 .. 59.999]        (seconds latitude/longitude)
    #        alt:    [-100000.00 .. 42849672.95] BY .01 (altitude in meters)
    #        siz, hp, vp: [0 .. 90000000.00] (size/precision in meters)
   def __init__(self, tokens):
        self.tokens=tokens
        # XXX todo : cop-out for now!

   def __str__(self):
       return string.join(self.tokens,' ')

class KeyFlags(Data):
#     0   1   2   3   4   5   6   7   8   9   0   1   2   3   4   5
#   +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
#   |  A/C  | Z | XT| Z | Z | NAMTYP| Z | Z | Z | Z |      SIG      |
#   +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
    Mnemonics={ 'NOKEY'  : 0xC000, 'NOCONF' : 0x8000,  'NOAUTH' : 0x4000, 'FLAG2'  : 0x2000,  'EXTEND' : 0x1000,
                'FLAG4'  : 0x0800, 'FLAG5'  : 0x0400,  'USER'   : 0x0000, 'ZONE'   : 0x0200,  'HOST'   : 0x0100,
                'NTYP3'  : 0x0300, 'FLAG8'  : 0x0080,  'FLAG9'  : 0x0040, 'FLAG10' : 0x0020,  'FLAG11' : 0x0010,
                'SIG15'  : 0x000F, 'SIG14'  : 0x000E,  'SIG13'  : 0x000D, 'SIG12'  : 0x000C,  'SIG11'  : 0x000B,
                'SIG10'  : 0x000A, 'SIG9'   : 0x0009,  'SIG8'   : 0x0008, 'SIG7'   : 0x0007,  'SIG6'   : 0x0006,
                'SIG5'   : 0x0005, 'SIG4'   : 0x0004,  'SIG3'   : 0x0003, 'SIG2'   : 0x0002,  'SIG1'   : 0x0001,
                'SIG0'   : 0x0000, 'SEP'    : 0x0001
              }
    
    def __init__(s, data):
        s.data = data
        try:
            s.flags = int(data)
        except ValueError:
            try:
                s.flags = s.mnem2num(data)
            except ValueError:
                s.flags = 0
        
    def mnem2num(s,data=None):
        if data==None:
            data=s.data
        s.flags=0
        for mn in data.split('|'):
            if not( KeyFlags.Mnemonics.has_key(mn) ):
                raise ValueError
            s.flags |= KeyFlags.Mnemonics[ mn ]
        return s.flags
        
    def num2mnem(s,flags=None):
        if flags==None:
            flags=s.flags
        d=''
        for mn in KeyFlags.Mnemonics.keys():
            if (KeyFlags.Mnemonics[ mn ] != 0) and (
               (flags & KeyFlags.Mnemonics[ mn ]) == KeyFlags.Mnemonics[ mn ]):
                d += mn + '|'
        if len(d):
            d = d[:-1]
        else:
            d = 'SIG0'
        return d
        
    def value(s):
        return s.flags
    
    def __str__(s):
        return s.num2mnem()

class KeyProtocol(Data):
    Mnemonics={ 'NONE'  : 0,
                'TLS'   : 1,
                'EMAIL' : 2,
                'DNSSEC': 3,
                'IPSEC' : 4,
                'ALL'   : 255
        }
    
    def __init__(s, data):
        s.data = data
        try:
            s.protocol = int(data)
        except ValueError:
            try:
                s.protocol = s.mnem2num(data)
            except ValueError:
                s.protocol = 0                

    def mnem2num(s,data=None):
        if data == None:
            data = s.data
        if data not in KeyProtocol.Mnemonics.keys():
            raise ValueError
        s.protocol = KeyProtocol.Mnemonics [ data ]
        
    def num2mnem(s,protocol=None):
        if protocol==None:
            protocol=s.protocol
        for mn in KeyProtocol.Mnemonics.keys():
            if protocol==KeyProtocol.Mnemonics[mn]:
                return mn
        return str(protocol)

    def value(self):
        return self.data
    
    def __str__(self):
        return self.num2mnem()
    
class KeyAlgorithm(Data):
    Mnemonics={ 'RSAMD5':      001,
                'DH':          002,
                'DSA':         003,   
                'ECC':         004,
                'RSASHA1':     005,                
                'INDIRECT':    252,    
                'PRIVATEDNS':  253,   
                'PRIVATEOID':  254
              }
    def __init__(s, data):
        s.data = data
        try:
            s.algorithm=int(data)
        except ValueError:
            try:
                s.algorithm=s.mnem2num(data)
            except ValueError:
                s.algorithm=0
    
    def mnem2num(s,data=None):
        if data==None:
            data=s.data
        if data in KeyAlgorithm.Mnemonics.keys() :
            return KeyAlgorithm.Mnemonics[ data ]
        raise ValueError

    def num2mnem(s,alg=None):
        if alg==None:
            alg=s.algorithm
        for mn in KeyAlgorithm.Mnemonics.keys() :
            if alg == KeyAlgorithm.Mnemonics[ mn ] :
                return mn
        return str(alg)

    def value(s):
        return str(s.algorithm)
    
    def __str__(s):
        return s.num2mnem()

class KeyTag(Data):
    def __init__(s,data):
        s.data=data
        try:
            s.tag = int(data)
        except ValueError:
            s.tag = 0
    
    def __str__(s):
        return str(s.tag)

class KeyDigest(Data):
    def __init__(s,data):
        s.data=data
        
    def __str__(s):
        return s.data
    
class CertType(Data):
    Mnemonics={ 'PKIX': 1,
                'SPKI': 2,
                'PGP' : 3,
                'URI' : 253,
                'OID' : 254
              }
    def __init__(s,data):
        s.data=data;
        try:
            s.type = int(data)
        except ValueError:
            try:
                s.type = s.mnem2num(data)
            except ValueError:
                s.type =0

    def mnem2num(s,data=None):
        if data==None:
            data=s.data
        if data in CertType.Mnemonics.keys() :
            return CertType.Mnemonics[ data ]
        return 0

    def num2mnem(s,data=None):
        if data==None:
            data=s.type
        for mn in CertType.Mnemonics.keys() :
            if data == CertType.Mnemonics[mn]:
                return mn
        return str(data)
    
    def value(s):
        return s.num2mnem()

    def str(s):
        return str(s.type)

class RRType(Data):
    def __init__(s, data):
        s.data = data

    def __str__(s):
        return s.data

class Timestamp(Data):
    def __init__(s,data):
        s.data=data
        
    def __str__(s):
        return s.data

class Base64(Data):
    def __init__(self,tokens):
        self.tokens=tokens
        
    def __str__(self):
        return string.join(self.tokens,'')

#    def value(s):
#        v =  string.join(s.tokens,'')
#        s.needMultilineBalancing = True
#        if ( len(v) > 32 ):
#            sv=v
#            v='( '
#            while len(sv):
#                v += sv[0:33] +'\n'
#                sv = sv[33:]
#            v += ' )'
#        return v
   
class List(Data):
    def __init__(self,tokens):
        self.tokens=tokens
        # XXX todo :cop-out for now!
    def __str__(self):
        return string.join(self.tokens,' ')

class Address(Data):
    def __init__(self,addr,allowPrefix=False):
        debug( 'DNS.Address',addr)
        if(addr.find(':')>-1):
            self.address = IPV6.Address(addr,IPV6.GUESS,allowPrefix)
        else:
            self.address = IPV4.Address(addr)

    def __str__(self):
        return str(self.address)

class AddressList(Data):
    def __init__(s,alist):
        s.input= alist
        s.addrs=[]
        debug( 'DNS.AddressList',alist)
        if alist != None:
            for a in alist.keys():
                s.addrs.append(Address(a))
                if (alist[a] != None) and (type(alist[a])==list) and (alist[a][0].lower()=='port'):
                    debug( '  PORT', alist[a][-1])
                    s.addrs[-1].port = int(alist[a][-1])                
        
    def __str__(s):
        out='{'
        for a in s.addrs:
            out += '\t'
            if hasattr(a,'port'):
                out += str(a)+' port '+str(a.port) + ';\n'
            else:
                out += str(a)+';\n'                
        out+='};\n'
        return out
        
class Path(Data):
    def __init__(self, path):
        self.path=path
        if( path == None ):
            self.path=''          
    
    def __str__(self):
        return self.path.strip('"')

    def value(self):
        return self.path

    def out(self):
        if (len(self.path)>1) and not( ( self.path[0] == '"') and (self.path[-1] == '"')):
            return '"'+self.path+'";'
        else:
            return self.path
 
class Boolean(Data):
    def __init__(self, val):
        if ((type(val)==int) and (val>0)) or ((type(val)==str) and (val.lower()=='yes') or (val=='1')):
            self.data = 1
        else:
            self.data = 0

    def __str__(s):
        if (s.data == None) or ( s.data <= 0):
            return 'no'
        else:
            return 'yes'
    
    def out(s):
        return str(s) +';'

class Radio(Data):
    def __init__(s,val,vals):
        if( type(val) == str ):
            i=0
            for v in vals:
                if v==val:
                    s.data = i
                i+=1
        else:
            s.data=val
        s.values = vals
        debug( 'DNS.RADIO: ',s.data,'VALS:',vals)
        
    def __str__(s):
        if type(s.data)==int:
            return s.values[s.data]
        else:
            return s.data

    def out(s):
        return str(s) + ';\n'


class Size(Data):
    def __init__(s,ss):
        s.scale=''
        s.sc=0
        if ss[-1] in ['K','k']:
            s.sc = 1024
            s.scale = 'K'
            ss=ss[:-1]
        elif ss[-1] in ['M', 'm']:
            s.sc = 1024**2
            s.scale = 'M'
            ss=ss[:-1]
        elif ss[-1] in ['G','g']:
            s.sc= 1024**3
            s.scale = 'G'
            ss=ss[:-1]
        if ss not in [ 'unlimited', 'default' ]:
            s.size=int(ss)
        else:
            s.size = ss
        debug( 'DNS.SIZE:',s.size,s.scale)
        
    def __str__(s):
        return str(s.size)+s.scale

class AnyAddress:
    def __init__(self):
        pass

    def __str__(s):
        return 'any'

class NoneAddress:
    def __init__(self):
        pass

    def __str__(s):
        return 'none'

class LocalAddress:
    def __init__(self):
        pass

    def __str__(s):
        return 'localhost'

class LocalNetsAddress:
    def __init__(self):
        pass

    def __str__(s):
        return 'localnets'

class Key(Data):
    name=''
    algorithm='hmac-md5'
    secret=''
    namedKeys={}
    def __init__(s,key,nmdc,name):        
        s.nmdc = nmdc        
        if name == None:
            return
        debug( 'NEW KEY: name:',name, ' key:', key)
        if type(name) == list:
            s.key = name
            s.name = name[-1].strip('"')
            if(  ((len(name) == 1) and (name[0] == 'key'))
              or ((len(name) == 3) and (name[0] == 'view'))              
              ):
                s.name = None
                debug("BAD KEY!")
                return
            if name[0] == 'view':
                s.nmKey=string.join([name[1],name[-1]],'_')
            else:
                s.nmKey=s.name
        else:
            s.key = [ 'key', name ]
            s.nmKey=name
            s.name = name.strip('"')
        s.nmdcKey = s.key
        debug( 'nmKey:', s.nmKey)
        if key == None:
            debug("NO KEY!")
            return
        s.name=s.name.strip('"')
        if Key.namedKeys.has_key(s.nmKey):
            s = Key.namedKeys[s.nmKey]
            debug( 'EXISTING KEY', s.nmKey,s.secret)
            return
        if key.has_key('algorithm'):
            s.algorithm = key['algorithm']
        if key.has_key('secret'):
            s.secret=key['secret'].strip('"')
        debug( 'CREATED NEW KEY:',s.name,s.nmKey,s.key, s.algorithm, s.secret)
        Key.namedKeys[s.nmKey] = s
    
    def __str__(s):
        if s.name == None:
            return ''
        return 'Key ' + s.name

    def value(s):
        return s.algorithm+ '; ' + s.secret +';'
    
    def out(s):
        return 'key'+ ' "' + s.name + '"' + '{\n\t' + 'algorithm ' + s.algorithm + ';\n\t' + 'secret "' + s.secret + '";\n};\n' 
        
class ACL(Data):
    namedACLs={}
    class Undefined(Exception):
        def __init__(s,acl,v):
            s.acl = acl
            s.name= v
                        
    def __init__(s,acls,nmdc,name=None,isControlKeys=False):
        s.addrs=[]
        s.acls=acls
        s.nmdc=nmdc
        s.notted = False
        s.key = name
        s.isControlKeys = isControlKeys
        if (type(s.key) == list) and (len(s.key)==2) and (s.key[0] == 'acl'):
            s.name = s.key[1]        
        else:
            s.name = None
        s.keyView = None
        if (type(s.key) == list) and (len(s.key)>2) and (s.key[0].lower() == 'view') and s.nmdc.__contains__( [s.key[0], s.key[1] ,'key']):
            s.keyView = [ s.key[0], s.key[1], 'key' ]
        debug( 'NEW ACL', s.key, s.name)
        if (s.name != None) and ACL.namedACLs.has_key(s.name):
            s = ACL.namedACLs[s.name]
            debug( 'EXISTING ACL:',s)
            return
        if (s.key != None) and len(s.key) and (type(s.key[-1])==str) and len(s.key[-1]) and (s.key[-1][0] == '!'):
            s.notted = True
        if( acls == None ):
            return # NEW 
        for v in acls.keys():
            notted = ( len(v) and  (v[0] == '!'))
            if( type(acls[v]) == dict ):
                s.addrs.append( ACL(acls[v],nmdc,s.key + [v]) )
                if notted:
                    s.addrs[-1].notted = True
                continue
            elif( (v != 'key') and (type(acls[v]) == list )):
                i=0
                for acl in acls[v]:
                    s.addrs.append( ACL(acl, nmdc, s.key + [v,i]))
                    if notted:
                        s.addrs[-1].notted = True
                    i+=1                    
                continue
            if notted:
                v = v[1:]
            if v == 'any':
                s.addrs.append(AnyAddress())
            elif v == 'none':
                s.addrs.append(NoneAddress())
            elif v == 'localhost':
                s.addrs.append(LocalAddress())
            elif v == 'localnets':
                s.addrs.append(LocalNetsAddress())
            elif v.find(':') > -1:                
                s.addrs.append(IPV6.Address(v,IPV6.GUESS,True))
            else:
                if isControlKeys:
                    if(s.nmdc.objs.has_key('key') and s.nmdc.objs['key'].has_key(v) ):
                        if not Key.namedKeys.has_key(v):
                            s.addrs.append( Key(s.nmdc.objs['key'][v],nmdc,v ) )
                        else:
                            s.addrs.append( Key.namedKeys[v] )
                    elif(s.nmdc.objs.has_key('key') and s.nmdc.objs['key'].has_key('"'+v+'"') ):
                        v='"'+v+'"'
                        if not Key.namedKeys.has_key(v):
                            s.addrs.append( Key(s.nmdc.objs['key'][v],nmdc,v ) )
                        else:
                            s.addrs.append( Key.namedKeys[v] )
                    else:
                        raise ACL.Undefined(s,v)                
                elif ( v == 'key' ):
                    nmdKeyKeys={}
                    for k in s.nmdc.objs['key'].keys():
                        nmdKeyKeys[ k.lower() ] = k
                    viewKeys = {}
                    if s.keyView != None:
                        for k in s.nmdc[s.keyView].keys():
                            nmdKeyKeys[ k.lower() ] = k
                            viewKeys[ k.lower() ] = k
                    ks=[]
                    if type(acls[v]) == list:
                        ks = acls[v];
                    else:
                        ks.append(acls[v])
                    for k in ks:
                        k = k.lower()
                        if not nmdKeyKeys.has_key( k ):
                            k = '"' + k + '"'
                            if not nmdKeyKeys.has_key( k ):
                                raise ACL.Undefined(s,v)
                        k = nmdKeyKeys[ k ] 
                        if not Key.namedKeys.has_key(k):
                            if viewKeys.has_key(k):
                                if not Key.namedKeys.has_key(s.keyView[1]+'_'+k):
                                    s.addrs.append( Key(s.nmdc[s.keyView + [ k ]],nmdc, s.nmdc[s.keyView + [ k ]] ) )
                                else:
                                    s.addrs.append( Key.namedKeys[ s.keyView[1]+'_'+k ])
                            else:
                                s.addrs.append( Key(s.nmdc.objs['key'][k],nmdc,k ) )
                        else:
                            s.addrs.append( Key.namedKeys[k] )
                else:
                    nmdACLkeys={}
                    if s.nmdc.objs.has_key('acl'):
                        for nmdak in s.nmdc.objs['acl'].keys():
                            nmdACLkeys[ nmdak.lower() ] = nmdak
                    if( s.nmdc.objs.has_key('acl') and  nmdACLkeys.has_key(v.lower()) ):
                        key = []
                        v = nmdACLkeys[ v.lower() ]
                        if( s.key != None):
                            key = s.key
                        if  not ACL.namedACLs.has_key(v):
                            ACL.namedACLs[v] = ACL(s.nmdc.objs['acl'][v],nmdc, ['acl',v])
                        s.addrs.append(ACL.namedACLs[v])
                    else:
                        a = IPV4.Address(v)
                        r = a.input(v,True)
                        if( r == None ):
                            s.addrs.append(a)
                        else:
                            raise ACL.Undefined(s,v)
            if notted:
                s.addrs[-1].notted = True
            if s.name != None:
                ACL.namedACLs[s.name] = s
    
    def out(s,t=''):
        out='{\n'
        for a in s.addrs:
            out +=t+"\t"
            if hasattr(a,'notted') and a.notted:
                out+='!'
            if( a.__class__ not in [ ACL, Key ] ):
                out += str(a)
                out +=';'
            else:
                if (a.__class__ == Key) and (not s.isControlKeys):
                    out += 'key '
                if a.name != None:
                    out += a.name+';'
                else:
                    out += a.out(t+"\t")
            out += '\n'                
        out+=t+'};'
        return out

    def __str__(s):
        if s.name == None:
            return 'ACL'
        else:
            name=''
            if type(s.name) == list:
                name=s.name[-1]
            else:
                name=s.name
            return 'ACL '+name
    
    def contains_acl(s,name):
        for a in s.addrs:
            if a.__class__== ACL:
                if a.name == name:
                    return True
                if a.contains_acl(name):
                    return True
        return False

    def contains_key(s,name):
        for a in s.addrs:
            if a.__class__== Key:
                if a.name == name:
                    return True
            elif a.__class__ == ACL:
                if a.contains_key(name):
                    return True
        return False

class SortList(ACL):
    def __init__(s,acls,nmdc,name=None):
        ACL.__init__(s,acls,nmdc,name)
        source=None
        order =None
        for i in range(0,len(s.addrs)):
            ele = s.addrs[i]
            if ele.__class__ != ACL:
                acl={ str(ele) : None };
                s.addrs[i] = ACL(acl,nmdc)
                ele = s.addrs[i]
            for j in range(0,min(2,len(ele.addrs))):
                if ele.addrs[j].__class__ != ACL:
                    acl={ str(ele.addrs[j]) : None };
                    ele.addrs[j] = ACL(acl,nmdc)
                debug( 'ELE ADDRS',i, ele.addrs[j].__class__)
            if len(ele.addrs) > 2:
                ele.addrs[1] += ele.addrs[2:]
                ele.addrs = ele.addrs[:2]
                debug( 'ELE ADDRS[1]', ele.addrs[1].__class__)
            debug( 'SORTLIST ELE:', ele.__class__, len(ele.addrs))

    def out(s):
        for acl in s.addrs:
            if (   (len(acl.addrs)==2) and (acl.addrs[0].__class__ == ACL)
               and (len(acl.addrs[0].addrs) == 1)
               and (acl.addrs[0].addrs[0].__class__ != ACL)
               ):
                acl.addrs[0] = acl.addrs[0].addrs[0]
        return ACL.out(s)
    
class Controls(Data):
    def __init__(s,nmdc):
        debug( 'NEW CONTROLS')
        s.inets=[]
        if not nmdc.__contains__('controls'):
            return
        controls = nmdc['controls']
        for inet in controls['inet'].keys():
            addr=inet[:-len(' allow')]
            port=None
            p=addr.find(' port')
            if p > -1:
                ps = addr[-p:]
                addr = addr[:p]
                port = int(ps[6:])
            address=IPV4.Address(addr)
            if port != None:
                address.port=port
            allow={}
            for k in controls['inet'][inet].keys():
                if k != 'keys':
                    allow[k]=None;
            s.inets.append(
                { 'addr':   address,
                  'keys':  ACL(controls['inet'][inet]['keys'],nmdc,name=None,isControlKeys=True),
                  'allow': ACL(allow,nmdc)                   
                })
    
    def out(s):
        c='controls {\n\t'
        for inet in s.inets:
            ads = str(inet['addr'])
            if hasattr(inet['addr'],'port') and (type(inet['addr'].port) != type(None)):                
                c += 'inet ' + ads + ' port ' + str(inet['addr'].port) + ' allow ' + inet['allow'].out()[:-1] + ' keys ' + inet['keys'].out() + '\n\t'
            else:
                c += 'inet ' + ads + ' allow ' + inet['allow'].out()[:-1] + ' keys ' + inet['keys'].out() + '\n\t'
        c=c[:-1]
        c+='};\n'
        return c

    def __str__(s):
        return s.out()

class Order:
    def __init__(self,order):
        self.order = order

class UpdatePolicy:
    def __init__(self,policy):
        self.policy = policy
    
class AlgorithmList:
    # XXX : todo
    pass

class PortList:
    # XXX : todo
    pass

class Logging:
    
    log_categories={
        'default':
        {  'desc':_('The default category defines the logging options for those categories where no specific configuration has been defined.')
        },
        'general':
        {  'desc':_('The catch-all. Many things still aren\'t classified into categories, and they all end up here.')
        },
        'database':
        {  'desc':_('Messages relating to the databases used internally by the name server to store zone and cache data.')        },
        'security':
        {  'desc':_('Approval and denial of requests.')
        },
        'config':
        {  'desc':_('Configuration file parsing and processing.')
        },
        'resolver':
        {  'desc':_('DNS resolution, such as the recursive lookups performed on behalf of clients by a caching name server.')
        },
        'xfer-in':
        {  'desc':_('Zone transfers the server is receiving.')
        },
        'xfer-out':
        {  'desc':_('Zone transfers the server is sending.')
        },
        'notify':
        {  'desc':_('The NOTIFY protocol.')
        },
        'client':
        {  'desc':_('Processing of client requests.')
        },
        'unmatched':
        {  'desc':_('Messages that named was unable to determine the class of or for which there was no matching view. A one line summary is also logged to the client category. This category is best sent to a file or stderr, by default it is sent to the null channel.')
        },
        'network':
        {  'desc':_('Network operations.')
        },
        'update':
        {  'desc':_('Dynamic updates.')
        },
        'queries':
        {  'desc':_('Queries. Using the category queries will enable query logging.')
        },
        'dispatch':
        {  'desc':_('Dispatching of incoming packets to the server modules where they are to be processed.')
        },
        'dnssec':
        {  'desc':_('DNSSEC and TSIG protocol processing.')
        },
        'lame-servers':
        {  'desc':_('Lame servers. These are misconfigurations in remote servers, discovered by BIND 9 when trying to query those servers during resolution.')
        },
        'delegation-only':
        {  'desc':_('Delegation only. Logs queries that have been forced to NXDOMAIN as the result of a delegation-only zone or a delegation-only in a hint or stub zone declaration.')
        }
    }
    syslog_facilities=[ 'auth', 'authpriv', 'cron', 'daemon', 'kern', 'lpr', 'mail', 'mark',
                        'news', 'security',  'syslog', 'user',  'uucp',
                        'local0','local1','local2','local3','local4','local5','local6','local7'
                      ]
    severities=[ 'critical', 'error', 'warning', 'notice', 'info', 'debug', 'level', 'dynamic' ]
    
    default_channels={ 'default_syslog':{ 'syslog': 'daemon',    'severity': 'info' },
                       'default_debug' :{ 'file':   'named.run', 'severity': 'dynamic' },
                       'default_stderr':{ 'stderr': None,        'severity': 'info'},
                       'null'          :{ 'null'  : None }
                     }
    
    default={'category': {'default': {'default_syslog': None,
                                        'default_debug': None
                                       },
                          'unmatched': {'null': None}
                         }
            }
    
    destinations=[ 'file', 'syslog', 'stderr', 'null' ]
    
    class File:
        def __init__(s, f):
            w = f
            if type(f) == list:
                w = f[:]
            elif( type(f) == type(None)):
                return
            else:
                w = f.split(' ')
            if len(w) == 0:
                return
            s.path = Path(w.pop(0))
            while len(w):
                o = w.pop(0)
                if o == 'versions':
                    if( len(w) ):
                        n = w.pop(0)
                        versions=n
                        if n.lower() != 'unlimited':
                            try:
                                versions = int(n)
                            except ValueError:
                                continue
                        s.versions = versions                        
                elif o == 'size':
                    if len(w):
                        s.size = Size(w.pop(0))
        
        def out(s):
            o=''
            if hasattr(s,'path'):
                o += s.path.out()
            else:
                return ''
            if hasattr(s,'versions'):
                o += ' ' + 'versions ' + str(s.versions)
            if hasattr(s,'size'):
                o += ' ' + 'size ' + s.size.out()
            return o
        
        def __str__(s):
            return s.out()
    
    class Channel:
        def __init__(s,name,ch):
            if(name==None) or (len(name)==0) or (type(ch) != dict):
                s.name = ''
                s.dest = 'null'
                s.severity=''
                s.opts={}
                return
            s.name = name.strip('"')
            s.opts={}
            if   ch.has_key('file'):
                s.file = Logging.File(ch['file'])
                s.dest = 'file'
            elif ch.has_key('syslog') and ( ch['syslog'].lower() in Logging.syslog_facilities ):
                s.syslog = ch['syslog']
                s.dest = 'syslog'
            elif ch.has_key('stderr'):
                s.dest = 'stderr'
                s.stderr=''
            elif ch.has_key('null'):
                s.dest ='null'
                s.null=''
            if not hasattr(s,'dest'):
                return
            sv=''
            if ch.has_key('severity'):
                severity=ch['severity']
                if type(severity)==list:
                    sv=severity[0].lower()
                else:
                    sv=severity.split()[0].lower()
            if  ch.has_key('severity') and (sv in Logging.severities):
                if type(severity) != list: 
                    sv=ch['severity'].split()
                else:
                    sv = severity
                if (len(sv)>1) and (sv[0].lower() == 'debug'):
                    dbg=None
                    try:
                        dbg=int(sv[1])
                    except ValueError:
                        dbg=None
                    if dbg != None:
                        s.severity = sv[0] + ' ' + sv[1]                        
                elif(len(sv)==1):
                    s.severity = ch['severity']
            if  ch.has_key('print-category'):
                s.opts['print-category']=ch['print-category']
            if ch.has_key('print-severity'):
                s.opts['print-severity']=ch['print-severity']
            if ch.has_key('print-time'):
                s.opts['print-time']=ch['print-time']
        
        def out(s):
            if not ( hasattr(s,'name') and hasattr(s,'dest') and ((s.dest in ['stderr','null']) or  hasattr(s, s.dest) )):
                debug( ' - DNS. LOGGING: BAD CHANNEL')
                return ''
            dest =''
            if not(s.dest in ['stderr','null']):
                dest=str(getattr(s,s.dest))
            o='\tchannel "'+s.name+'" {\n\t\t'+s.dest +' '+dest+';\n'
            if hasattr(s,'severity') and len(s.severity) and (s.severity in Logging.severities) :
                o+='\t\tseverity'+' '+s.severity+';\n'
            for opt in s.opts.keys():
                o+='\t\t'+opt+' '+s.opts[opt]+';\n'
            o+='\t};\n'
            return o
    
    class Category:
        def __init__(s,logging,name,cat):
            name=name.strip('"')
            if name not in Logging.log_categories.keys():
                debug( 'UNKNOWN CATEGORY:',name)
                return
            s.name=name
            s.channels=[]
            channels = logging.channels.keys()  + Logging.default_channels.keys()
            for channel in cat.keys():
                channel=channel.strip('"')
                if channel not in channels:
                    debug( 'CATEGORY: UNKNOWN CHANNEL:',name,channel)
                    s.channels = []
                    return
                s.channels.append(channel)            
            
        def out(s):
            if (not (hasattr(s,'channels') and hasattr(s,'name'))) or (len(s.channels)==0):
                return ''
            o='\tcategory '+s.name+' {\n'
            for ch in s.channels:
                o +='\t\t'+ ch +';\n'
            o += '\t};\n'
            return o
    
    def __init__(s,nmdc):
        s.channels={}
        s.categories=[]
        s.nmdc = nmdc
        if nmdc.__contains__('logging'):
            logging = s.nmdc['logging']
        else:
            logging={}
        s.parse(logging)
    
    def parse(s, lg):
        for c in Logging.default_channels.keys():
            s.channels[c] = Logging.Channel(c,Logging.default_channels[c])
        if lg.has_key('channel'):
            for c in lg['channel'].keys():
                s.channels[c.strip('"')] = Logging.Channel(c,lg['channel'][c])
        hasDefault=False
        hasUnmatched=False
        if lg.has_key('category'):
            for c in lg['category'].keys():
                if c == 'default':
                    hasDefault=True
                if c == 'unmatched':
                    hasUnmatched=True
                s.categories.append(Logging.Category(s,c,lg['category'][c]))            
        if not hasDefault :
            s.categories.append(Logging.Category(s,'default',Logging.default['category']['default']))
        if not hasUnmatched:
            s.categories.append(Logging.Category(s,'unmatched',Logging.default['category']['unmatched']))
        
    def out(s):
        defaultChannels=Logging.default_channels.keys()        
        o = 'logging {\n'
        for c in s.channels.keys():
            if c not in defaultChannels:
                o += s.channels[c].out()
        for c in s.categories:
            if c.name=='default':
                printDefault=False
                for ch in c.channels:
                    if ch not in ['default_syslog','default_debug']:
                        printDefault=True
                        break
                if printDefault:
                    o += c.out()
            elif c.name=='unmatched':
                printDefault=False
                if ch in c.channels:
                    if ch != 'null':
                        printDefault=True
                        break
                if printDefault:
                    o+=c.out()
            else:
                o += c.out()
        o += '};\n'
        return o
