#!/usr/bin/python
# Hosts file parser for system-config-bind Import facility
#   <ip address> <canonical name> <aliases...>
# Jason Vas Dias, Red Hat Inc., Oct 2004
import sys
import __builtin__
import gettext
__builtin__.__dict__['_'] = gettext.gettext
import DNS

class Filter:
    AND=1
    OR=2
    def filter(s,h):
        return True

    def __str__(s):
        return _('( any )')

    def attr(s):
        return ''

    def desc(s):
        return _('Host Name Suffix Filter')
    
class IPV4_Filter(Filter):
    def __init__(s,ip):
        if( ip.__class__ != DNS.IPV4.Address ):
            s.ip = DNS.IPV4.Address(ip,input=True,prefixAllowed=True)
        else:
            s.ip = ip
        t=s.ip.prefix
        s.addr=s.ip.int()
        s.mask=((2**32)-1)
        if (s.ip.prefix != None):
            s.mask &= ~(( 2**(32-s.ip.prefix))-1)
        debug( 'IPV4:', '%x' % s.addr, 'MASK:', '%x' % s.mask)
        
    def filter(s,h):
        ip=DNS.IPV4.Address(h,input=True)
        if len(ip.addr)<=0:
            return False or (hasattr(s, 'notted') and s.notted)
        a=ip.int()
        if hasattr(s, 'notted') and s.notted:
            return (a & s.mask) != (s.addr & s.mask)
        return (a & s.mask) == (s.addr & s.mask)

    def __str__(s):
        if hasattr(s,'notted') and s.notted:
            return '!' + str(s.ip)
        return str(s.ip)
    
    def attr(s):
        return s.ip

    def desc(s):
        return _('Host IPV4 Address Filter')

    
class IPV6_Filter(Filter):
    def __init__(s,ip):
        if(ip.__class__ != DNS.IPV6.Address):
            s.ip = DNS.IPV6.Address(ip,allow_prefix=True)
        else:
            s.ip = ip
        s.addr = s.ip.int()
        s.mask = (2**128)-1
        if( s.ip.prefix != None):
            s.mask &= ~(( 2**(128-s.ip.prefix))-1)
        debug( 'IPV6:', '%x' % s.addr, 'MASK:', '%x' % s.mask)
        
    def filter(s,h):
        ip = DNS.IPV6.Address(h)
        if ip.address == None:
            return False or (hasattr(s, 'notted') and s.notted)
        a=ip.int()
        if hasattr(s, 'notted') and s.notted:
            return (a & s.mask) != (s.addr & s.mask)
        return (a & s.mask) == (s.addr & s.mask)

    def __str__(s):
        if hasattr(s,'notted') and s.notted:
            return '!' + str(s.ip)
        return str(s.ip)
    
    def attr(s):
        return s.ip

    def desc(s):
        return _('Host IPV6 Address Filter')

    
class Name_Filter(Filter):
    def __init__(s,name):
        s.name = name

    def filter(s,n):
        if hasattr(s,'notted') and s.notted:
            return (n[-len(s.name):] != s.name)
        return (n[-len(s.name):] == s.name)

    def __str__(s):
        if hasattr(s,'notted') and s.notted:
            return '!' + str(s.name)
        return str(s.name)
    
    def attr(s):
        return s.name
    
class Hosts:
    def parse(s, filename=None, namesToo=False):
        if( filename == None ):
            filename = s.filename;
        s.hosts={}
        s.names={}
        try:
            hosts=file(filename,"r")
        except IOError,e:
            return
        for line in hosts:
            line=line.rstrip()
            i=line.find("#")
            if( i > -1 ):
                line=line[0:i]
            line=line.strip()
            if( len(line) > 0 ):
                tokens=line.split()
                if( len(tokens) > 1):
                    if( not s.hosts.has_key(tokens[0])):
                        if( len(tokens) > 2):
                            s.hosts[tokens[0]] = [ tokens[1], tokens[2:] ]
                        else:
                            s.hosts[tokens[0]] = [ tokens[1], [] ]
                    else:
                        for name in tokens[1:]:
                            if not( (name == s.hosts[tokens[0]][0]) or (name in s.hosts[tokens[0]][1])):
                                s.hosts[tokens[0]][1].append( name )
                    if namesToo:
                        for name in tokens[1:]:
                            if not s.names.has_key( name ):
                                s.names[name]=[tokens[0],[]]
                            elif not( (tokens[0]==s.names[name][0]) or (tokens[0] in s.names[name][1])):
                                s.names[name][1].append( tokens[0] )
        hosts.close()
        
    def __init__(s, filename, namesToo=False):
        if (filename == None) or (len(filename) == 0):
            filename='/etc/hosts'
        s.filename = filename        
        s.parse(namesToo=namesToo)
        bad_file=False
        for ip in s.hosts.keys():
            ipv4 = DNS.IPV4.Address(ip,input=True)
            if len(ipv4.addr) == 0:
                ipv6 = DNS.IPV6.Address(ip)
                if( ipv6.address == None ):                    
                    bad_file=True
                    break
        if bad_file :
            s.hosts={}        
        s.filters = []

    def add_filter(s,f):
        if issubclass( f.__class__, Filter ):
            s.filters.append(f)
            return s.filters[-1]
        if DNS.IPV4.Address.ip4_prefix_re.match(f):
            ip = DNS.IPV4.Address(f,input=True,prefixAllowed=True)
            s.filters.append(IPV4_Filter(ip))
        else:
            ip = DNS.IPV6.Address(f,allow_prefix=True)
            if ip.address != None:
                s.filters.append(IPV6_Filter(ip))                
            else:
                s.filters.append(Name_Filter(f))
        if len(s.filters):
            return s.filters[-1]
        return None
    
    def filter(s,array,filters=None,op=Filter.OR):
        if( array == None):
            return
        if(filters == None):
            filters=s.filters
        for ip in s.hosts.keys():
            match=not(len(s.filters)>0)
            for filter in filters:                                    
                if (filter.__class__ == Name_Filter):
                   if( filter.filter(s.hosts[ip][0]) ):
                       match=True
                   elif op == Filter.AND:
                       match=False
                       break
                else:
                   if( filter.filter(ip) ):
                       match=True
                   elif op == Filter.AND:
                       match=False
                       break 
            if match:                                
                 array.append([ip, s.hosts[ip][0], s.hosts[ip][1]])

    def keys(s):
        return s.hosts.keys()
    
    def n_hosts(s):
        a=s.hosts.keys()
        return len(a)

if __name__ == '__main__':
    file=None
    if( len(sys.argv)>=2 ):
        file=sys.argv[1]
    h=Hosts(file)
    if( len(sys.argv)>2):
        for host in h.filter(sys.argv[2].split(',')):
            debug( host)




