#
# BIND OSI NSAP address classes
#
# Jason Vas Dias <jvdias@redhat.com>, Red Hat Inc. Sep 2004

import DNS

REVERSE=-4
REVERSE_ZONE='NSAP.INT.'

class Address(DNS.Data):
    addr=''
    address=''

    def strip(s, addrIn=None):
        if addrIn == None:
            addr = s.addr[:]
        else:
            addr = addrIn[:]
        if addr == None:
            return ''
        a=''
        if (len(addr) >= 2) and (addr[:2]=='0x'):
            addr=addr[2:]
        for c in addr[:]:
            if not c in '.0123456789ABCDEFabcdef':
                debug( '   NSAP BAD CHAR: ',c)
                return None
            if c == '.':
                continue
            a+=c
        return a
            
    def __init__(s, addr):
        if( addr == None):
            s.addr=''
            s.address=''
            return
        s.addr=addr        
        s.address=s.strip(addr)            
        
    def __str__(self):
        if self.addr == None:
            return ''
        return  self.label()
    
    def section(s, b, e, addr=None):
        if addr == None:
            addr = s.address
        l = len(addr)
        if   l > (e+1):
            return addr [ b : e+1 ] + '.'
        elif l == (e+1):
            return addr [ b : e+1 ]
        elif l >= b:
            return addr [ b : ]
        else:
            return ''
    
    def remaining_sections(s,e,addr=None):
        if addr == None:
            addr = s.address
        a=''
        i=e
        l=len(addr)
        if l > i:
            a = '.'
        while i < l:
            a += s.section(i,i+4,addr)
            i += 4                
        return a

    # Convention for 1st 20 octets of variable length NSAP address (RFC 1706):
    #              |-----|--------|-------------------------------------|
    #              | 47  |  0005  | DFI | AA |Rsvd | RD |Area | ID |Sel |
    #              |-----|--------|-----|----|-----|----|-----|----|----|
    #       octets |  1  |   2    |  1  | 3  |  2  | 2  |  2  | 6  | 1  |
    #              |-----|--------|-----|----|-----|----|-----|----|----|
    
    def label(s):
        if (s.address == None) or (len(s.address)<=0):
            return ''
        return ( '0x' + s.section(0,1) 
                      + s.section(2,5)
                      + s.section(6,7)
                      + s.section(8,13)
                      + s.section(14,17)
                      + s.section(18,21)
                      + s.section(22,25)
                      + s.section(26,37)
                      + s.section(38,39)
                      + s.remaining_sections(40)
               )
    
    def value(s):
        if s.address == None:
            return ''
        return s.label()
    
    
class ReverseAddress( Address ):    
    raddr=''
    origin=''
    zone_origin=''
    
    def reverse(s,addr=None,with_dots=False):
        if(addr == None):
            addr = s.raddr[:]
        if addr == None:
            return None
        if( (len(addr) >= len(REVERSE_ZONE)) and
            (addr[-(len(REVERSE_ZONE)):].upper() == REVERSE_ZONE)
          ):
            addr = addr[:-(len(REVERSE_ZONE)+1)]
        addr = s.strip(addr)
        ra=''
        l=len(addr)
        for i in range(l-1,-1,-1):
            c=addr[i]
            ra += c
            if with_dots and (i > 0):
                ra += '.'
        return ra                
    
    def __init__(self, addr, zone):
        self.raddr=addr
        self.origin=zone.origin
        self.zone_origin = zone.zone_origin
        Address.__init__(self,self.reverse(str(self)))
        self.origin_address = Address( self.reverse(self.zone_origin) )
        
    def __str__(self):
        if self.raddr == '@':
            return self.zone_origin
        if len(self.raddr) and (self.raddr[-1] == '.'):
            return self.raddr
        elif self.zone_origin == '.':
            return self.raddr + self.zone_origin
        else:
           return self.raddr + '.' + self.zone_origin

    def value(self):
        if self.raddr == self.zone_origin:
            return '@'
        return self.raddr

    def relative_label(self):
        a = self.label()
        o = self.origin_address.label()
        debug( 'NSAP relative label:', a,o)
        if(   (len(a) >= len(o))
           and(a[:len(o)].upper() == o.upper())
          ):
            if( len(a) > len(o) ) and ( a[len(o)] == '.' ):                
                return a[( len(o) + 1):]
            else:
                return a[len(o):]
        return a
    
    def contains(s, a):
        addr=s.strip(a)
        if(    (len(addr) >= len(s.origin_address.address))
           and (addr[:len(s.origin_address.address)] ==  s.origin_address.address)
          ):
            debug( 'NSAP CONTAINS: ',addr, s.origin_address.address)
            return True
        debug( 'NSAP CONTAINS NOT: ',addr, s.origin_address.address)
        return False
    
    def split_domain(s):
        a = s.strip(s.reverse(str(s)))
        o = ''
        for l in range(26,2,-4):
            if len(s.address > l):
                o = s.address[0:l]
                a = s.address[l:]
                break
        if (len(o)>0) :
            return ( s.reverse(a) , s.reverse(o)+'.'+REVERSE_ZONE )
        elif(len(a)>0):
            return ( s.reverse(a), REVERSE_ZONE )
        else:
            return ( '',  REVERSE_ZONE  )

    def relative_value(self,origin=None):
        if origin == None:
            if has_attr(self,'zone_origin'):
                origin = self.zone_origin
            else:
                origin = ''
        s = self.__str__()
        if (len(origin)>0) and (len(s) >= len(origin)) and (s[-len(origin):]==origin):
            s=s[:-len(origin)]
            if( len(s) == 0):
                s = '@'
            if(len(s) > 1) and (s[-1] == '.'):
                s=s[:-1]
        return s

def reverse(ip):
    if((len(ip)>=2) and (ip[0:2]=='0x')):
        ip = ip[2:]
    if((len(ip)>=len(REVERSE_ZONE)) and (ip[-len(REVERSE_ZONE):] == REVERSE_ZONE)):
        ip = ip[:-(len(REVERSE_ZONE)+1)]
    a=''
    for c in ip:
        if c in '0123456789ABCDEFabcdef':
            a=c+'.'+a
    if(len(a)>0):
        a=a[:-1]
    return a
